/*
part of resizer-yuv.c
this code is taken from spuc/generic/butterworth.h
SPUC - Signal processing using C++ - A DSP library
Copyright(c) 1993-1998 Tony Kirke

resizer-yuv is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.
  
resizer-yuv is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
  
You should have received a copy of the GNU General Public License
along with GNU Make; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "resizer.h"

#ifdef USE_SPUC_CODE

long order;
long odd;
long n2;
complex<double>* roots;
iir_2nd< Numeric>* iir;
iir_1st< Numeric>* iir_1;
double gain;
double* coeff;

// Constructor, fcd = cut-off (1=sampling rate)
// ord = Filter order
// amax = attenuation at cut-off

butterworth(double fcd, long ord, double amax)
{
double epi;
double wca;
double roots;

// amax - attenuation at cutoff
ord  = 1;
amax = 3.0;

gain = 1;
order = ord;
epi = pow( (pow(10.0, (amax / 10.) ) - 1.0) , (1. / (2.0 * ord) ) );
// fcd - desired cutoff frequency (normalized)
wca = 2.0 * tan(M_PI * fcd) / epi;
// wca - pre-warped angular frequency
n2 = (order + 1) / 2;
odd = (order % 2);
roots = new complex<double> [n2];
if (odd) iir_1 = new iir_1st< Numeric >;
get_roots(wca, order, n2);
bilinear(n2);
iir = new iir_2nd < Numeric > [order / 2];
get_coeff(n2);
} /* end function butterworth */


void bilinear(long n2)
{
double td;
int j;
const double a = 2.;

if(odd)
	{
	/*
	For s-domain zero (or pole) at infinity we
	get z-domain zero (or pole) at z = -1
	*/

	double tmp = (roots[0].real() - a) / (a + roots[0].real());
	iir_1 -> set_coeff(-tmp);
	gain *= 0.5 * (1 + tmp);
	}
for(j = odd; j < n2; j++)
	{                       
	td = a * a - 2 * a * roots[j].real() + magsq(roots[j]);
	roots[j] = complex<double>((a * a - magsq(roots[j])),
	2.0 * a * roots[j].imag()) / td;
	}
} /* end function bilinear */


void get_coeff(long n2)
{
for (int j = odd;j < n2;j++)
	{ 
	iir[j - odd].set_coeff(-2 * roots[j].real() / magsq(roots[j]), 1.0 /\
	 magsq(roots[j]) );

	gain *= \
		(magsq(roots[j]) - 2 * roots[j].real() + 1.0)/(4 * magsq(roots[j])); 
	}
} /* end function get_coeff */


void get_roots(double wp, long n, long n2)
{
double arg;

long l = 0;
if(n % 2 == 0) l = 1;                           

for (int j = 0;j < n2; j++)
	{
	arg = -0.5 * M_PI * l / float(n);
	roots[j] = wp * expj(arg);
	l += 2;
	}
} /* end function get_roots */


Numeric clock(Numeric in)
{
Numeric tmp = in;
for(int i = odd; i < n2; i++)
	{
	tmp = iir[i - odd].clock(tmp);
	}

if (odd) tmp = iir_1 -> clock(tmp);

return(gain * tmp);
} /* end function clock */


// Destructor
~butterworth()
{
delete [] roots;
if (odd) delete iir_1;
delete [] iir;
} /* end function ~butterworth */


// Reset history
void reset()
{
for (int j = odd; j < n2; j++)
	{
	iir[j - odd].reset();
	if (odd) iir_1 -> reset();
	}
} /* end function reset */


// Unit magnitude polar to rectangular conversion
inline complex<double> expj(double x)
{
complex<double> y;

y.re = cos(x);
y.im = sin(x);
return y;
} /* function complex */

#endif

/*
Type : biquad, tweaked butterworth
References :Posted by Patrice Tarrabia
Code :
r  = rez amount, from sqrt(2) to ~ 0.1
f  = cutoff frequency
 (from ~0 Hz to SampleRate/2 - though many
 synths seem to filter only  up to SampleRate/4)

The filter algo:
out(n) = \
	a1 * in + a2 * in(n-1) + a3 * in(n-2) - b1 * out(n-1) -  b2 * out(n-2)

Lowpass:
 c = 1.0 / tan(pi * f / sample_rate);
 a1 = 1.0 / ( 1.0 + r * c + c * c);
 a2 = 2 * a1;
 a3 = a1;
 b1 = 2.0 * ( 1.0 - c * c) * a1;
 b2 = ( 1.0 - r * c + c * c) * a1;

Hipass:
 c = tan(pi * f / sample_rate);
 a1 = 1.0 / ( 1.0 + r * c + c * c);
 a2 = -2 * a1;
 a3 = a1;
 b1 = 2.0 * ( c * c - 1.0) * a1;
 b2 = ( 1.0 - r * c + c * c) * a1;
*/

