/*
resizer-yuv.c

Copyright (C) Jan Panteltje  2002

resizer-yuv is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.
  
resizer-yuv is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
  
You should have received a copy of the GNU General Public License
along with GNU Make; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "resizer.h"

//#include "fftw.h"
#include <rfftw.h>

/* active part of a TV line, 50 uS */
#define PAL_ACTIVE_LINE .00005

fftw_real *in, *out;

/* maximum size of Kaiser window */
double pkd[10000];

int verbose_flag;

rfftw_plan plan_luma, plan_chroma, plan_reverse_luma, plan_reverse_chroma;

void help_print()
{
fprintf(stderr,\
"\
Usage: resizer-yuv [-b beta] [-c] [-d] [-e end level] [-h] [-k]\n\
[-l low pass mode] [-m] [-r resize mode] [-s start level] [-v]\n\
[-w window size] [-x width] [-y height].\n\n\
\t-b  (float) beta for Kaiser window used in sinc resizing (default 7.5).\n\
\t-c  reverse u and v (switch colors).\n\
\t-d  print debug messages (o.a. function calls and arguments).\n\
\t-e  (float) end level, attenuation at niquist point f / 2 (default .1).\n\
\t-h  print command line options (this help), and exit.\n\
\t-l  (int) low pass mode, 0=off, 1=luminance only, 2=YUV, 3=frame (default 3).\n\
\t-m  fft measure, does first test runs to find best algo (takes a while).\n\
\t-r  (int) resizer mode, 0=off, 1=sinc over frame, 2=sinc YUV, 3=resizer (default 3).\n\
\t-s  (float) start level, default .74, attenuation at 70%% of niquist point f / 2.\n\
\t-v  verbose, print Kaiser window and, filter frequency versus attenuation in dB.\n\
\t-w  (int) window size for sinc function in resampling (default 40).\n\
\t-x  (int) new width in pixels.\n\
\t-y  (int) new height in pixels.\n\
");
} /* end function help_print */


/* taken from spuc */
double io(double x)
{
/*
function:  io
Bessel function for Kaiser window
*/
const double t = 1.e-08;
double y = 0.5 * x;
double e = 1.0;
double de = 1.0;
int i;
double xi;
double sde;

for(i = 1; i < 26; i++)
	{
	xi = (double)i;
	de *= y / xi;
	sde = de * de;
	e += sde;
	if( (e * t - sde) > 0) break;
	}

//fprintf(stderr, "WAS io x=%.2f returning e=%.2f\n", x, e);
return e;  
} /* end function io */


/* taken from spuc, corrected by Jan Panteltje */
void kaiser(double *w, long nf, double beta)
{
/*
Kaiser window
nf = filter length in samples
beta = parameter of Kaiser window
*/
double io(double b);
double bes = 1.0 / io(beta);
double xind = (nf - 1) * (nf - 1);
int i;
int odd;
double xi;
double da;

odd = nf % 2; //?????? not in original!
for(i = 0; i < nf; i++)
	{
	if(odd) xi = i + 0.5;
	else xi = i;
	xi = 4 * xi * xi;

	da = 1. -xi / xind;
	if(da < 0) da = 0;

	w[i] = io(beta * sqrt(da) ) * bes;
	}

if(verbose_flag)
	{
	for(i = 0; i < nf / 2; i++)
		{
		fprintf(stderr,\
		"resizer-yuv(): kaiser window w[%d]=%.5f\n", i, w[i]);
		}
	} /* end if verbose_flag */
} /* end function kaiser */


double hs(double t, double Fs)
/*
Formula from Yehar's digital sound processing tutorial for the braindead!
version 2001.05.02M
www.student.oulu.fi/~oniemita
Coded to C by Jan Panteltje
*/   
{
register double da, db;

db = M_PI * Fs * t; // already radians * ( (2 * M_PI) / 360.0);

da =  sinl(db) / db;

return da;
} /* end function hs */


static double get_value_at_t(\
	double t, double Ts, unsigned char ar[], int points,\
	int window)
/*
Formula from Yehar's digital sound processing tutorial for the braindead!
version 2001.05.02M
www.student.oulu.fi/~oniemita
Coded to C by Jan Panteltje
*/   
/*
x(t) = sum for n = -00 to n = +00 x(n.Ts).hs(t - n.Ts)
where hs = sin(pi.Fs.t) / pi.Fs.t
Ts = sampling period
Fs is sampling frequency
n ranges over integers
t = time in seconds (real)
*/
{
register int n;
register int pos, a, min, max;
register double Fs;
register double da, db, dc, dd, de, df, dy;

Fs = 1.0 / Ts;

/* position */
da = t / Ts;
pos = (int)da;

df = M_PI * Fs;

a = window / 2;
min = pos - a;
max = pos + a;

dy = 0.0;
for(n = min; n < max; n++)
	{
	if(n < 0) continue;

	if(n > points) continue;

	db = pkd[ abs(pos - n) ];

//db = 1.0;

	/* function hs() inline for speed */
	dc = t - ( (double)n * Ts);
	de = df * dc; // M_PI * Fs * dc;
	dd = sin(de) / de;

	dy += (double)ar[n] * dd * db;

//	dy += (double)ar[n] * hs(t - ( (double)n * Ts), Fs) * db;

	} /* end for n */

return dy;
} /* end function get_value_at_t */ 


//void pre_calculate_coeficients(int points, double coef[])
//{
//} /* end function pre_calculate_coeficients */


char *strsave(char *s)
/* save char array s somewhere */
{
char *p;

if(p = malloc( strlen(s) +  1) ) strcpy(p, s);
return(p);
} /* end function strsave */


int main(int argc, char **argv)
{
int a, c, i, j, z;
char temp[4096];
double da, db, dc, dd, de, df;
static int x, y, u, v, y1, r, g, b;
int idy, new_idy;
uint8_t *pino, *pouto;
uint8_t *py, *pu, *pv;
uint8_t *pyo, *puo, *pvo;
uint8_t uya, uyb, uyc, uyt1, uyt2;
static int cr, cg, cb, cy, cu, cv;
int interlace_mode;
double aspect_ratio;
int length, ylength, ulength, vlength;
int new_length, new_ylength, new_ulength, new_vlength;
char *ptr;
FILE *outfile;
char *bptr;
int width, height, frame_rate1, frame_rate2;
int new_width, new_height;
int current_frame_nr;
char metadata[READSIZE];
int aspect_ratio1, aspect_ratio2;
int new_aspect_ratio1, new_aspect_ratio2;
char *running;                                             
char *token;
int reverse_uv_flag;
int fft_measure_flag;
int points, new_points;
unsigned char *inbuffer, *outbuffer;
unsigned char *os_line_buffer;
double *dmultiplier;
int filter_order;
double dhor_scale_factor, dvert_scale_factor;
int start_x, end_x;
double start_level, end_level;
int window_size;
int low_pass_mode;
int resize_mode;
double beta;

/* identify */
fprintf(stderr, "\nPanteltje (c) resizer-yuv%s\n\n", VERSION);

/* get home directory */  
userinfo = getpwuid(getuid() );
home_dir = strsave(userinfo -> pw_dir);
user_name = strsave(userinfo -> pw_name);

/* set some defaults */
debug_flag = 0;

/* for stdin to stdout */
outfile = stdout;

reverse_uv_flag = 1;
fft_measure_flag = 0;
filter_order = 40;
start_level = 0.7412;
end_level = 0.1;
/*
40 - 7.5, 20 - 3.6, 10 - 2.3
*/
window_size = 40;
beta = 7.5; // The bigger the smaller the Kaiser window
low_pass_mode = 3; // frame
resize_mode = 3; // transcode resizer
/* end defaults */

/* test for command line argumenst */
if(argc == 1)
	{
	fprintf(stderr,\
	"resizer-yuv(): no arguments\n");

	help_print();

	exit(1);
	} /* end if argc == 1 */

/* get command line options, these override the defaults */
while(1)
	{
	a = getopt(argc, argv, "b:cde:hl:mr:s:vw:x:y:");
	if(a == -1) break;
	switch(a)
		{
		case 'b':
			beta = atoi(optarg);
			break;
		case 'c':
			reverse_uv_flag = 1 - reverse_uv_flag;
			break;
        case 'd':
			debug_flag = 1;
			break;
		case 'e':
			end_level = atof(optarg);
			break;
		case 'h':
			help_print();
			exit(0);	
			break;
		case 'l':
			low_pass_mode = atoi(optarg);
			if(low_pass_mode > 2)
				{
				fprintf(stderr,\
				"resizer-yuv(): unsupported low pass mode %d\n",\
				low_pass_mode);

				exit(1);
				}
			break;
		case 'm':
			fft_measure_flag = 1 - fft_measure_flag;
			break;
		case 'r':
			resize_mode = atoi(optarg);
			if(resize_mode > 3)
				{
				fprintf(stderr,\
				"resizer-yuv(): unsupported resize mode %d\n",\
				resize_mode);

				exit(1);
				}
			break;
		case 's':
			start_level = atof(optarg);
			break;
		case 'v':
			verbose_flag = 1 - verbose_flag;
			break;
		case 'w':
			window_size = atoi(optarg);
			break;
		case 'x':
			new_width = atoi(optarg);
			break;
		case 'y':
			new_height = atoi(optarg);
			break;
		default:
			help_print();
			exit(1);
			break;
		}/* end switch a */
	} /* end while getopt() */

if(debug_flag)
	{
	fprintf(stderr,\
	"command line parser result: new_width=%d new_height=%d filter_order=%d\n",\
	new_width, new_height, filter_order);
	}

if( (resize_mode == 1) || (resize_mode == 2) )
	{
	kaiser(pkd, window_size, beta);
	}

/* read yuv stream from stdin */
/*
From mjpegtools:
Description of the (new!, forever?) YUV4MPEG2 stream format:

STREAM consists of
 o one '\n' terminated STREAM-HEADER
 o unlimited number of FRAMEs

FRAME consists of
 o one '\n' terminated FRAME-HEADER
 o "length" octets of planar YCrCb 4:2:0 image data
   (if frame is interlaced, then the two fields are interleaved)

STREAM-HEADER consists of
 o string "YUV4MPEG2 "  (note the space after the '2')
 o unlimited number of ' ' separated TAGGED-FIELDs
 o '\n' line terminator

FRAME-HEADER consists of
 o string "FRAME "  (note the space after the 'E')
 o unlimited number of ' ' separated TAGGED-FIELDs
 o '\n' line terminator

TAGGED-FIELD consists of
 o single ascii character tag
 o VALUE (which does not contain whitespace)

VALUE consists of
 o integer (base 10 ascii representation)
 or o RATIO
 or o single ascii character
 or o generic ascii string

RATIO consists of
 o numerator (integer)
 o ':' (a colon)
 o denominator (integer)

The currently supported tags for the STREAM-HEADER:
	 W - [integer] frame width, pixels, should be > 0
 H - [integer] frame height, pixels, should be > 0
 I - [char] interlacing:  p - progressive (none)
                          t - top-field-first
                          b - bottom-field-first
                          ? - unknown
 F - [ratio] frame-rate, 0:0 == unknown
 A - [ratio] sample (pixel) aspect ratio, 0:0 == unknown
 X - [character string] 'metadata' (unparsed, but passed around)

The currently supported tags for the FRAME-HEADER:
 X - character string 'metadata' (unparsed, but passed around)
*/

/* stream stream header to get width and height and interlace */

/*
YUV4MPEG2 W720 H576 F25:1 I? A0:0 XM2AR002
FRAME
data
*/

/* create a line buffer */
bptr = malloc(READSIZE);
if(! bptr)
	{
	fprintf(stderr, "resizer-yuv(): malloc bptr failed, aborting.\n");

	exit(1);
	}

#ifndef NO_GNU_C
/*
From libc.info:
ssize_t getline (char **LINEPTR, size_t *N, FILE *STREAM)
This function is a GNU extension, but it is the recommended way to
read lines from a stream.  The alternative standard functions are 
unreliable.
*/

b = READSIZE;
a = getline(&bptr, &b, stdin);
if(a == -1)
	{
	fprintf(stderr,\
	"resizer-yuv(): EOF in read stream header, stop.\n");

	exit(0);
	}
#else
/*
code for non GNU systems contributed by "Steven M. Schultz" <sms@2BSD.COM>
*/
for(b = 0; b < READSIZE - 1; b++)
	{
	bptr[b] = fgetc(stdin);
	if(feof(stdin))
		{
		fprintf(stderr, "resizer-yuv: EOF in read stream header, stop.\n");

		exit(0);
		}

	if (bptr[b] == '\n') break;
	}
/* keep the \n */
bptr[b + 1] = '\0';

if(b == READSIZE - 1)
	{
	fprintf(stderr,\
	"resizer-yuv: No newline in first %d bytes of stream header\n",\
	READSIZE);

	exit(1);
	}
#endif

if(strncmp(bptr, "YUV4MPEG2", 9) != 0)
	{
	fprintf(stderr,\
	"resizer-yuv(): YUV4MPEG2 not found in stream header, aborting.\n");

	exit(1);
	}

if(debug_flag)
	{
	fprintf(stderr, "incoming stream header=%s\n", bptr);
	}

/* parse arguments */
running = strsave(bptr);
if(! running)
	{
	fprintf(stderr,\
	"resizer-yuv(): strsave(options) failed, aborting.\n");

	exit(1);
	}

while(1)
	{
	token = strsep (&running, " ");
	if(token == NULL) break;

	/* avoid empty string */
	if(token[0] == 0) continue;
	sscanf(token, "A%d:%d", &aspect_ratio1, &aspect_ratio2);
	sscanf(token, "W%d", &width);
	sscanf(token, "H%d", &height);
	sscanf(token, "F%d:%d", &frame_rate1, &frame_rate2);
	sscanf(token, "I%c", &interlace_mode);
	sscanf(token, "X%s", metadata);
	} /* end while parse tags */

free(running);

if(debug_flag)
	{
	fprintf(stderr,\
	"width=%d height=%d\n\
	frame_rate1=%d frame_rate2=%d\n\
	interlace_mode=%c\n\
	aspect_ratio=%.2f\n",
	width, height,\
	frame_rate1, frame_rate2,\
	interlace_mode,\
	aspect_ratio);
	}

if(! resize_mode)
	{
	new_width = width;
	new_height = height;
	}

/* keep aspect */
new_aspect_ratio1 = aspect_ratio1;
new_aspect_ratio2 = aspect_ratio2;

/* calculate length */
ylength = width * height;
ulength = (width / 2) * (height / 2);
vlength = (width / 2) * (height / 2);
length = ylength + ulength + vlength;

/* calculate new_length */
new_ylength = new_width * new_height;
new_ulength = (new_width / 2) * (new_height / 2);
new_vlength = (new_width / 2) * (new_height / 2);
new_length =  new_ylength + new_ulength + new_vlength;

/* create plans for fft and reverse fft */
if(fft_measure_flag)
	{
	fprintf(stderr,\
	"resizer-yuv(): testing fft algos for best speed, please wait.\n");

	a = FFTW_MEASURE;
	}
else a = 0;

if( (low_pass_mode == 1) || (low_pass_mode == 2) )
	{
	/* creates plan for forward fft */
	plan_luma = rfftw_create_plan(\
		width, FFTW_REAL_TO_COMPLEX, FFTW_OUT_OF_PLACE | a);
	fprintf(stderr, "resizer-yuv(): have forward luma plan.\n");

	/* create plan for reverse fft */
	plan_reverse_luma = rfftw_create_plan(\
		width, FFTW_COMPLEX_TO_REAL, FFTW_OUT_OF_PLACE | a);
	fprintf(stderr, "resizer-yuv(): have reverse luma plan.\n");	

	in = malloc( width * sizeof(fftw_real) );
	if(! in)
		{
		fprintf(stderr, "resizer-yuv(): malloc in[] failed, aborting\n");

		exit(1);
		}

	out = malloc( width * sizeof(fftw_real) );
	if(! out)
		{
		fprintf(stderr, "resizer-yuv(): malloc out[] failed, aborting\n");

		exit(1);
		}

	dmultiplier = malloc( (width / 2) * sizeof(double) );
	if(! dmultiplier)
		{
		fprintf(stderr,\
		"resizer-yuv(): malloc dmultiplier failed, aborting\n");

		exit(1);
		}
	}

if(low_pass_mode == 2)
	{
	/* creates plan for forward fft color */
	plan_chroma = rfftw_create_plan(\
		width / 2, FFTW_REAL_TO_COMPLEX, FFTW_OUT_OF_PLACE | a);
	fprintf(stderr, "resizer-yuv(): have forward chroma plan.\n");

	/* create plan for reverse fft color */
	plan_reverse_chroma = rfftw_create_plan(\
		width / 2, FFTW_COMPLEX_TO_REAL, FFTW_OUT_OF_PLACE | a);
	fprintf(stderr, "resizer-yuv(): have reverse chroma plan.\n");
	}
if(low_pass_mode == 3)
	{
	/* creates plan for forward fft */
	plan_luma = rfftw_create_plan(\
		length, FFTW_REAL_TO_COMPLEX, FFTW_OUT_OF_PLACE | a);
	fprintf(stderr, "resizer-yuv(): have forward luma plan.\n");

	/* create plan for reverse fft */
	plan_reverse_luma = rfftw_create_plan(\
		length, FFTW_COMPLEX_TO_REAL, FFTW_OUT_OF_PLACE | a);
	fprintf(stderr, "resizer-yuv(): have reverse luma plan.\n");	

	in = malloc( length * sizeof(fftw_real) );
	if(! in)
		{
		fprintf(stderr, "resizer-yuv(): malloc in[] failed, aborting\n");

		exit(1);
		}

	out = malloc( length * sizeof(fftw_real) );
	if(! out)
		{
		fprintf(stderr, "resizer-yuv(): malloc out[] failed, aborting\n");

		exit(1);
		}

	dmultiplier = malloc( (length / 2) * sizeof(double) );
	if(! dmultiplier)
		{
		fprintf(stderr,\
		"resizer-yuv(): malloc dmultiplier failed, aborting\n");

		exit(1);
		}
	}

if(! resize_mode)
	{
	fprintf(stderr,\
	"resizer-yuv(): Found - r 0 flag, no resizing done!\n");
	}
else if(resize_mode == 3)
	{
	/* setup the transcode re-sizer */
	bb_resize_setup(width, height, new_width, new_height, debug_flag);
	}

if(! low_pass_mode)
	{
	fprintf(stderr,\
	"resizer-yuv(): Found -l 0 flag, no lowpass filter applied!\n");
	}

inbuffer = malloc(length);
if(! inbuffer)
	{
	fprintf(stderr, "resizer-yuv(): malloc inbuffer failed, aborting\n");

	exit(1);
	}

outbuffer = malloc(new_length);
if(! outbuffer)
	{
	fprintf(stderr, "resizer-yuv(): malloc outbuffer failed, aborting\n");

	exit(1);
	}

if(debug_flag)
	{
	fprintf(stderr,\
	"main(): calculated length=%d new_length=%d\n",\
	length, new_length);

	fprintf(stderr, \
	"inbuffer=%p outbuffer=%p\n", inbuffer, outbuffer);
	}

/* set pointers */
py = inbuffer;   
pyo = outbuffer;

if(! reverse_uv_flag)
	{
	pu = inbuffer + width * height;
	pv = inbuffer + (width * height * 5) / 4;

	puo = outbuffer + new_width * new_height;
	pvo = outbuffer + (new_width * new_height * 5) / 4;
	}
else
	{
	pv = inbuffer + width * height;
	pu = inbuffer + (width * height * 5) / 4;

	pvo = outbuffer + new_width * new_height;
	puo = outbuffer + (new_width * new_height * 5) / 4;
	}	

/* send new stream header */
fprintf(outfile,\
"YUV4MPEG2 W%d H%d F%d:%d I%c A%d:%d X%s\n",\
new_width, new_height, frame_rate1, frame_rate2, interlace_mode,\
new_aspect_ratio1, new_aspect_ratio2, metadata);

if( (low_pass_mode == 1) || (low_pass_mode == 2) || (low_pass_mode == 3) )
	{
	/* 
	pre-calculate  an array of values that will be used to modify the
	frequency spectrum
	*/

	/*
	gaussian curve 
	From <javaplus@hotmail.com> in sci.math:
	Say your spectrum is F(u,v) and you
	want to remove point F(u0,v0) without introducing
	ripples. Multiply F(u,v) by W(u,v) where W(u,v) is
	equal 1 for all (u,v) except a small window centered
	at (x0,y0). Call this region G(u,v) and say it is 7x7.
	If you choose G(u,v)=0 then you get ripples.
	But if you use a simple separable Gaussian
	then you should not notice any ripples.
	In this case G(u,v)=g(u).g(v) where

	1 - exp(-u^2 / a)

	where a = 4 for a 7 x 7 window.

	g(3)=0;
	g(2)=g(4)=0.2212 = 1 - e^(- 1 x 1 / 4)
	g(1)=g(5)=0.6321 = 1 - e^(- 2 x 2 / 4)
	g(0)=g(6)=0.8946 = 1 - e^(- 3 x 3 / 4)
	*/	

	/*
	Some values for exp(x)
	y = exp(x):
	x         y
	-5    --> .0067 
	-4    --> .018
	-3    --> .049
	-2.3  --> .1
	-2	  --> .13
	-1    --> .3
	-.5   --> .6
	-.25  --> .778
	-.125 --> .88
	*/

	/* scale factor */
	if( (low_pass_mode == 1) || (low_pass_mode == 2) ) points = width / 2;
	else if(low_pass_mode == 3) points = length / 2;

	dhor_scale_factor = (double)new_width / (double)width;  
	dvert_scale_factor = (double)new_height / (double)height;

	/* end point slope and star point slope */
	end_x = points * dhor_scale_factor;
	start_x = (double)points * dhor_scale_factor * .7;

	fprintf(stderr, "resizer_yuv(): start_x=%d end_x=%d\n", start_x, end_x);

	/* calculate value of b and filter order for this filter at f / 2 */ 
	filter_order = 10; /* start somewhere */
	while(1)
		{
		/* start somewhere */
		if( (low_pass_mode == 1) || (low_pass_mode == 2) ) de = 1000.0;
		if(low_pass_mode == 3) de = (double)length * 2.0;

		a = points - end_x;
		/* here to cut off */

		while(1)
			{
			db = ( (double)a * (double)a) / de;
			da = exp( -db);
			da = 1.0 - da;

			/* number of lowpasses (order) */
			dc = pow(da, filter_order);
			if(dc <= end_level) break;
	
			de += (double)a / 4.0;
//fprintf(stderr,\
//"WAS length=%d a=%d da=%.6f db=%.6f dc=%.6f de=%.6f\n",\
//length, a, da, db, dc, de);

			} /* end while increment b */

		if(debug_flag)
			{
			fprintf(stderr, "using filter_order=%d de=%.6f\n",\
			filter_order, de);
			}

		/* test start slope */
		a = points - start_x;

		db = ( (double)a * (double)a) / de;
		da = exp( -db);

		da = 1.0 - da;
	
		/* number of lowpasses (order) */
		dc = pow(da, filter_order);
	
		if(dc > start_level) break; // .73

		filter_order++;
		} /* end while increment order */

	fprintf(stderr,\
	"resizer-yuv(): using filter_order=%d de=%.2f\n",\
	filter_order, de);

	for(x = 0; x < points; x++)
		{
		/* 1 - exp(-u^2 / a) */

		/* points in curve */
		a = points - x;

		db = ( (double)a * (double)a) / de;
		da = exp( -db);
		da = 1.0 - da;

		/* number of lowpasses (order) */
		dc = pow(da, filter_order);

		dmultiplier[x] = dc;

		/* freq in MHz */
		dd = (double)x / (double)points * 7.2;

		/* calculate dB for report */
		db = 10 * log(dmultiplier[x]);

		if(verbose_flag)
			{
			fprintf(stderr,\
			"resizer_yuv(): MHz=%.2f point=%d multiplier=%.4f %.2fdB\n",\
			dd, x, dmultiplier[x], db); 
			}
		} /* end for x (all points) */

	/* prevent dc shift if filter active at zero (color error) */
	dmultiplier[0] = 1.0;
	dmultiplier[1] = 1.0;
	} /* end if low_pass_mode */

/* read frames in YUV format from stdin */
current_frame_nr = 1;
while(1)
	{
	/* read frame header */
#ifndef NO_GNU_C
	b = READSIZE;
	a = getline(&bptr, &b, stdin);
	if(a == -1)
		{
		fprintf(stderr,\
		"resizer-yuv(): EOF in read frame header, stop.\n");
		
		exit(0);
		}
#else
	/*
	code for non GNU systems contributed by Steven M. Schultz
	*/
	for(b = 0; b < READSIZE - 1; b++)
		{
		bptr[b] = fgetc(stdin);
		if(feof(stdin))
			{
			fprintf(stderr,\
			"resizer-yuv: EOF in read frame header, stop.\n");

			exit(0);
			}

		if (bptr[b] == '\n') break;
		}
	/* keep the /n */
	bptr[b + 1] = '\0';

	if(b == READSIZE - 1)
		{
		fprintf(stderr,\
		"resizer-yuv: No newline in first %d bytes of frame header, aborting.\n",\
		READSIZE);

		exit(1);
		}
#endif

	if(strncmp(bptr, "FRAME", 5) != 0)
		{
		fprintf(stderr,\
		"resizer-yuv(): No FRAME found in frame header, aborting.\n");

		exit(1);
		}

	/* 
	This is how it is written:
	Write Y', Cb, and Cr planes
	for (i = 0; i < 3; i++)
		{
		unsigned char *srctop = upper_field[i];
		unsigned char *srcbot = lower_field[i];
		// alternately write one line from each
		for (y = 0; y < height; y += 2)
			{
			if (y4m_write(fd, srctop, width)) return Y4M_ERR_SYSTEM;
			srctop += width;
			if (y4m_write(fd, srcbot, width)) return Y4M_ERR_SYSTEM;
			srcbot += width;
			}

		// for chroma, width/height are half as big
		if (i == 0)
			{
			width /= 2;
			height /= 2;
			}
		}
	*/

	/* read in frame data */
	a = fread(inbuffer, sizeof(char), length, stdin);
	if(a != length)
		{
		fprintf(stderr,\
		"resizer-yuv(): unexpected EOF in frame data, aborting.\n");

		exit(1);
		}

	if(debug_flag)
		{
		fprintf(stderr, "main(): have frame=%d\n", current_frame_nr);
		}

	if(low_pass_mode == 3)
		{
		/* fft over whole data */
		points = length;
		/* get data in right format */
		for(i = 0; i < points; i++)
			{
			in[i] = (double)inbuffer[i];
			} /* end for i */

		/* forward fft */
		rfftw_one(plan_luma, in, out);

		/* modify frequency spectrum */
		for(i = 0; i < points / 2; i++)
			{
			/*
			fft result real part in first half array,
			imaginary part in second half array.
			*/

			/* real part */
			out[i] *= dmultiplier[i];

			/* complex part */
			out[points - i] *= dmultiplier[i];
			} /* end for all i (frequency components) */

		/* reverse fft */
		rfftw_one(plan_reverse_luma, out, in);

		/* data back to right format */
		for(i = 0; i < points; i++)
			{
			da = in[i];

			/* correct amplitude */
			da /= points;

//#define TEST_RANGE_MODE
#ifdef TEST_RANGE_MODE
			if(da > 255) fprintf(stderr, "WAS da=%.2f\n", da);
			if(da < 0) fprintf(stderr, "WAS da=%.2f\n", da);
#endif TEST_RANGE_MODE

			/* limit range, sometimes this is < 0 and sometimes > 255! */
			if(da > 255) da = 255;
			if(da < 0) da = 0;

			py[i] = (unsigned char)da;
			} /* end for i*/
		} /* end if low_pass_mode == 3 */

	if( (low_pass_mode == 1) || (low_pass_mode == 2) )
		{
		/* fft over a line */

		/* Y luminance low pass */
		points = width;
		for(y = 0; y < height; y++)
			{
			idy = y * width;

			/* get data in right format */
			for(x = 0; x < points; x++)
				{
				in[x] = (double)py[idy + x];
				} /* end for x */

			/* forward fft */
			rfftw_one(plan_luma, in, out);

			/* modify frequency spectrum */
			for(x = 0; x < points / 2; x++)
				{
				/*
				fft result real part in first half array,
				imaginary part in second half array.
				*/

				/* real part */
				out[x] *= dmultiplier[x];

				/* complex part */
				out[points - x] *= dmultiplier[x];
				} /* end for all x (frequency components) */

			/* reverse fft */
			rfftw_one(plan_reverse_luma, out, in);

			/* data back to right format */
			for(x = 0; x < points; x++)
				{
				da = in[x];

				/* correct amplitude */
				da /= points;

//#define TEST_RANGE_MODE
#ifdef TEST_RANGE_MODE
				if(da > 255) fprintf(stderr, "WAS da=%.2f\n", da);
				if(da < 0) fprintf(stderr, "WAS da=%.2f\n", da);
#endif TEST_RANGE_MODE

				/* limit range, sometimes this is < 0 and sometimes > 255! */
				if(da > 255) da = 255;
				if(da < 0) da = 0;

				py[idy + x] = (unsigned char)da;
				} /* end for x */
			} /* end for y */
		} /* end if low_pass_mode */

	if(low_pass_mode == 2)
		{
		/* U chrominance low pass */
		points = width / 2;
		for(y = 0; y < height / 2; y++)
			{
			idy = y * points;

			/* get data in right format */
			for(x = 0; x < points; x++)
				{
				in[x] = (double) pu[idy + x];
				} /* end for x */
	
			/* forward fft */
			rfftw_one(plan_chroma, in, out);

			/* mask out high frequencies */
			for(x = 0; x < points / 2; x++)
				{
				/* real part */
				da = out[x] - 127;
				da *= dmultiplier[x * 2];
				out[x] = da + 127; 

				/* complex part */
				da = out[points - x] - 127;
				da *= dmultiplier[x * 2];
				out[points - x] = da + 127;
				} /* end for x */

			/* reverse fft */
			rfftw_one(plan_reverse_chroma, out, in);

			/* data back to right format */
			for(x = 0; x < points; x++)
				{
				da = in[x];
	
				/* correct amplitude */
				da /= points;

				/* limit range */
				if(da > 255) da = 255;
				if(da < 0) da = 0;
//da = 127;
				pu[idy + x] = (unsigned char)da;
				} /* end for x */
			} /* end for y */

		/* V chrominance low pass */
		points = width / 2;
		for(y = 0; y < height / 2; y++)
			{
			idy = y * points;

			/* get data in right format */
			for(x = 0; x < points; x++)
				{
				/* mask out burst area if specified */
				in[x] = (double)pv[idy + x];
				} /* end for x */

			/* forward fft */
			rfftw_one(plan_chroma, in, out);

			/* mask out high frequencies */
			for(x = 0; x < points / 2; x++)
				{
				/* real part */
				da = out[x] - 127;
				da *= dmultiplier[x * 2];
				out[x] = da + 127; 

				/* complex part */
				da = out[points - x] - 127;
				da *= dmultiplier[x * 2];
				out[points - x] = da + 127;
				} /* end for x */

			/* reverse fft */
			rfftw_one(plan_reverse_chroma, out, in);

			/* data back to right format */
			for(x = 0; x < points; x++)
				{
				da = in[x];

				/* correct amplitude */
				da /= points;

				/* limit range */
				if(da > 255) da = 255;
				if(da < 0) da = 0;

//da = 127;
				pv[idy + x] = (unsigned char)da;
				} /* end for x */
			} /* end for y */
		} /* end if low_pass_mode == 2 */ 

	/* data now is in inbuffer */

output:
	if(resize_mode == 1)
		{
		/* sinc interpolation over frame *\
		/*
		We should use PAL frame time of 40 mS (.04) but this gives math
		errors, resulting in several spickels in a frame.
		So we normalize to 1.0, as it only about the ratios, dd now no
		longer holds true time though.
		*/
#define NORMALIZER 1.0

		/* input sample time */
		da = NORMALIZER / (double)width;

		/* output sample time */
		db = NORMALIZER / (double)new_width;

		for(i = 0; i < new_length; i++)
			{
			dd = (double)i * db;

			dc = get_value_at_t(dd, da, inbuffer, length, window_size);

//#define TEST_RANGE_MODE
#ifdef TEST_RANGE_MODE
			if(dc > 255) fprintf(stderr, "WAS dc=%.2f\n", dc);
			if(dc < 0) fprintf(stderr, "WAS dc=%.2f\n", dc);
#endif TEST_RANGE_MODE

			/* limit range, sometimes this is < 0 and sometimes > 255! */
			if(dc > 255) dc = 255;
			if(dc < 0) dc = 0;

			outbuffer[i] = (unsigned char)dc;
			} /* end for i */
		} /* end if resize_mode == 1 */
	else if(resize_mode == 2)
		{
		/* sinc interpolation over yuv separately */ 
		/*
		50 uS per line,
		input sample time is .00005 / width
		output sampe time is .00005 / new_width 
		*/

		/* input sample time */
		da = PAL_ACTIVE_LINE / (double)width;

		/* output sample time */
		db = PAL_ACTIVE_LINE / (double)new_width;

		/* Y */
		points = width;
		new_points = new_width;

		/* for all lines in screen */
		for(y = 0; y < height; y++)
			{
			idy = y * points;
			new_idy = y * new_points;

			for(x = 0; x < new_width; x++)
				{
				/* time t = x * output sample time */
				dd = (double)x * db;
//fprintf(stderr,\
//"WAS x=%d tsample1=dd=%.16f inputssmplt=%.16f outputsmplt=%.16f\n",\
//x, dd, da, db);
				/*
				double get_value_at_t(\
				double t, double tsample1, unsigned char ar[],\
				int points, int window)
				*/
				dc = \
				get_value_at_t(dd, da, &py[idy], points, window_size);

				/* mask out start */					
				if(x < 3) dc = 0.0;

				pyo[new_idy + x] = (unsigned char)dc;
				} /* end for x */
			} /* end for y */

		/* U */
		points = width / 2;
		new_points = new_width / 2;
		for(y = 0; y < height / 2; y++)
			{
			idy = y * points;
			new_idy = y * new_points;

			for(x = 0; x < new_width / 2; x++)
				{
				/* time t = x * output sample time */
				dd = (double)x * db;

				dc = \
				get_value_at_t(dd, da, &pu[idy], points, window_size / 2);

				/* mask out start */
				if(x < 3) dc = 127.0;
//dc = 127.0;
				puo[new_idy + x] = (unsigned char)dc;

				} /* end for x */
			} /* end for y */

		/* V */
		points = width / 2;
		new_points = new_width / 2;
		for(y = 0; y < height / 2; y++)
			{
			idy = y * points;
			new_idy = y * new_points;

			for(x = 0; x < new_width / 2; x++)
				{
				/* time t = x * output sample time */
				dd = (double)x * db;

				dc = \
				get_value_at_t(dd, da, &pv[idy], points, window_size  / 2);

				/* mask out start */
				if(x < 3) dc = 127.0;
//dc = 127;
				pvo[new_idy + x] = (unsigned char)dc;
				} /* end for x */
			} /* end for y */
		} /* end if resize_mode == 2 */
	else if(resize_mode == 3)
		{
		bb_resize_frame(py, pu, pv, pyo, puo, pvo);
		}

	/* echo frame header */
	fprintf(outfile, "%s", bptr);

	/* write the frame to the output */
	if(! resize_mode)
		{
		a = fwrite(inbuffer, sizeof(char), new_length, outfile);
		}
	else
		{
		a = fwrite(outbuffer, sizeof(char), new_length, outfile);
		}
	if(a != new_length)
		{
		fprintf(stderr,\
		"resizer-yuv(): fwrite(): failed,\n\
		tried to write %d bytes, only could write %d bytes, aborting.\n",\
		new_length, a);

		exit(1);
		}

	current_frame_nr++;
	} /* end while frames */

/* never here! */
exit(1);
} /* end function main */

